''*******************************************************
''*  TV Driver with minimal hub memory use  v2.0        *
''*  Author: Chip Gracey                                *
''*  Modifications by: Michael Green                    *
''*  Copyright (c) 2004 Parallax, Inc.                  *
''*  Modifications copyright (c) 2010 by Michael Green  *
''*  See end of file for terms of use.                  *               
''*******************************************************

' v1.0 - 01 May 2006 - original version
' v1.1 - 17 May 2006 - pixel tile size can now be 16 x 32 to enable more
'                      efficient character displays utilizing the internal
'                      font.  See 'tv_mode'
' v2.0 - 12 Jan 2010 - rewritten based on Michael Park's driver that places
'                      the display buffer in the cog's memory.  The user
'                      interface has changed to allow increased functionality.
'                      The current display row (1-13) is in interface.byte[3].
'                      The current display column (1-40) is in interface.byte[2]
'                      The interface.byte[1] is a command code and
'                      interface.byte[0] is an operand for the command.
'                      The low order 4 or 6 bits of the command may be a repeat
'                      count.  If repeat count is zero, nothing is done
'                      Commands: $00     - Idle
'                                $01     - Copy display char to operand, C++
'                                $02     - Disable display driver I/O pins
'                                $03     - Enable display driver I/O pins
'                                $04     - Set foreground color from operand
'                                $05     - Set background color from operand
'                                $06-$0F - Reserved (ignored)
'                                $10-$1F - Fill current row with operand, R++
'                                $20-$2F - Move next row upwards, R++
'                                $30-$3F - Move previous row downwards, R--
'                                $40-$7F - Output operand to display, C++
'                                $80-$BF - Move next character to left, C++
'                                $C0-$FF - Move previous character to right, C--
'                      Note: The current display column and row are updated
'                            after each occurance of an operation as noted.
'                            The command byte is zeroed when completed.
'                      Note: This is expected to provide a 40 x 13 display.
'                            There's an additional blank line above and below
'                            the display area to facilitate scrolling.  Row 0 is
'                            the blank row above the display area and row 14 is
'                            the blank row below the display area.  Column 0 is
'                            the last column of the previous row and column 41
'                            is the first column of the next row.

CON
  cols          = 40            'number of columns on a display line
  rows          = 13            'number of display lines on the display

  #$00                          'ANSI Control Codes
  NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS,  HT,  LF,  VT,  FF,  CR,  SO,  SI
  DLE, DC1, DC2, DC3, DC4, NAK, SYN, ETB, CAN, EM, xSUB, ESC, FS,  GS,  RS,  US
  #$7F
  DEL

  Bsp           = BS
  Tab           = HT
  
  doNothing     = $00           'do nothing
  getDispChar   = $01           'copy display character to operand, C++
  disableDrvr   = $02           'disable display driver I/O pins
  enableDrvr    = $03           'enable display driver I/O pins
  setForeClr    = $04           'set foreground color from operand
  setBackClr    = $05           'set background color from operand
  fillRow       = $10           'fill current row with operand, R++
  moveRowUp     = $20           'move next row upwards, R++
  moveRowDown   = $30           'move previous row downwards, R--
  putDispChar   = $40           'output operand to display, C++
  deleteLeft    = $80           'move next character to left, C++
  deleteRight   = $C0           'move previous character to right, C--

VAR
   long saveRendezvous, escape

PUB start(rendezvous)
   saveRendezvous := rendezvous
   escape~
   
PUB str(s)                                              'display string at s
   repeat strsize(s)
      out(byte[s++])

PUB dec(v) | i, x                                       'display decimal number
   x := v == NEGX                                       'check for max negative
   if v < 0
      v := ||(v+x)                                      'adjust as needed
      out("-")                                          ' and output sign
   i := 1_000_000_000                                   'initialize divisor
   repeat 10                                            'loop for 10 digits
      if v => i                                                               
         out(v / i + "0" + x*(i == 1))                  'display digit
         v //= i
         result~~                                       'flag non-zero found
      elseif result or i == 1
         out("0")                                       'display at least zero
      i /= 10

PUB hex(v,w)                                            'display hexadecimal
   v <<= (8 - w) << 2                                   ' at most 8 digits
   repeat w
      out(lookupz((v <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(v, w)                                           'display binary
   v <<= 32 - w                                         ' at most 32 digits
   repeat w
     out((v <-= 1) & 1 + "0")

PUB out(c) | size                                       'display single char
   case escape
      0:    case c
               BS:    if getDispCol                     'already at column 1?
                         byte[saveRendezvous+2]--       'if not, back up,
                         sendToCog(putDispChar+1," ")   ' write a space, then
                         waitForCog                     '  back up again
                         byte[saveRendezvous+2]--
               HT:    repeat                            'display spaces until
                         sendToCog(putDispChar+1," ")   ' next multiple of 8
                         checkNewline                   '  (minimum 1 space)
                      while getDispCol & 7
               CR:    if size := cols - getDispCol      'fill line with spaces
                         sendToCog(putDispChar+size," ")
                      checkNewLine                      ' then do a new line
               DC1, DC2, ESC:
                      escape := c
               other: sendToCog(putDispChar+1,c)
                      checkNewline
      DC1:  waitForCog
            byte[saveRendezvous+2] := 1 <# c <# cols    'column is 1..40
            escape~
      DC2:  waitForCog
            byte[saveRendezvous+3] := 1 <# c <# rows    'row is 1..13
            escape~
      ESC:  sendToCog(putDispChar+1,c)                  'escaped character
            checkNewLine
            escape~

PRI waitForCog
   repeat while byte[saveRendezvous+1]                  'wait for idle state

PRI sendToCog(command,char)
   waitForCog                                           'wait for idle state
   word[saveRendezvous] := command<<8 + char            'set new command to do

PRI getDispCol                                          'return column in range 0-39
   waitForCog
   return byte[saveRendezvous+2] - 1

PRI checkNewline                                        'check for end of line
   if getDispCol == cols
      byte[saveRendezvous+2] := 1
      if ++byte[saveRendezvous+3] > rows                'roll up and blank fill
         long[saveRendezvous] := constant(1<<24+1<<16+(moveRowUp+rows)<<8)
         waitForCog                                     'wait for roll up
         long[saveRendezvous] := constant(rows<<24+1<<16) 'set correct row

{{
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}}
